from typing import Dict, List, NamedTuple, Optional, Set, Tuple, Union
import appdirs
import bz2
import json
from pathlib import Path
import requests
import logging
import sys
import unicodedata

logger = logging.Logger("yajwiz")

DATA_DIR = Path(appdirs.user_data_dir("yajwiz"))
DICTIONARY_PATH = DATA_DIR / "dictionary.json"

KAWHAQ_URL = "https://de7vid.github.io/qawHaq/"
FORMAT = "iOS-1"

def _parse_boqwiz_pos(pos: str) -> Tuple[str, Set[str]]:
    colon = pos.split(":")
    if len(colon) == 1:
        return colon[0], set(colon)
    
    [tpos, tags] = colon
    comma = tags.split(",")
    return tpos, set([tpos]+comma)

class BoqwizEntry(NamedTuple):
    """
    Represents an entry in the boQwI' dictionary, describing a single headword.
    """

    id: str
    name: str
    part_of_speech: str
    simple_pos: str
    tags: Set[str]
    definition: Dict[str, str]
    notes: Dict[str, str]
    examples: Dict[str, str]
    search_tags: Dict[str, List[str]]
    hidden_notes: Optional[str]
    synonyms: Optional[str]
    antonyms: Optional[str]
    see_also: Optional[str]
    components: Optional[str]
    source: Optional[str]

    @staticmethod
    def from_json(id: str, data: dict) -> "BoqwizEntry":
        tpos, tags = _parse_boqwiz_pos(data["part_of_speech"])
        return BoqwizEntry(
            id=id,
            name=data["entry_name"],
            part_of_speech=data["part_of_speech"],
            simple_pos=tpos,
            tags=tags,
            definition=data["definition"],
            notes=data.get("notes", {}),
            examples=data.get("examples", {}),
            search_tags=data.get("search_tags", {}),
            hidden_notes=data.get("hidden_notes", None),
            synonyms=data.get("synonyms", None),
            antonyms=data.get("antonyms", None),
            see_also=data.get("see_also", None),
            components=data.get("components", None),
            source=data.get("source", None)
        )

class BoqwizDictionary(NamedTuple):
    """
    Represents the whole boQwI' dictionary.
    """

    version: str
    locales: Dict[str, str]
    supported_locales: List[str]
    entries: Dict[str, BoqwizEntry]

    @staticmethod
    def from_json(data: dict) -> "BoqwizDictionary":
        return BoqwizDictionary(
            version=data["version"],
            locales=data["locales"],
            supported_locales=data["supported_locales"],
            entries={
                key: BoqwizEntry.from_json(key, value)
                for key, value in data["qawHaq"].items()
            }
        )
    
    def __repr__(self):
        return f"<BoqwizDictionary version={self.version}>"

def _normalize(data: Union[str, int, float, dict, list]) -> Union[str, int, float, dict, list]:
    if isinstance(data, str):
        return unicodedata.normalize("NFC", data)
    
    elif isinstance(data, dict):
        return {
            _normalize(key): _normalize(value)
            for key, value in data.items()
        }
    
    elif isinstance(data, list):
        return [_normalize(item) for item in data]
    
    else:
        return data

def _try_load() -> Optional[dict]:
    try:
        DATA_DIR.mkdir(parents=True, exist_ok=True)
        if DICTIONARY_PATH.exists():
            with open(DICTIONARY_PATH, "r") as f:
                data = f.read()
                data = json.loads(data)
                data = _normalize(data)
            
            return data
        
        else:
            return None
    
    except:
        logger.error("Error while reading the dictionary!", exc_info=sys.exc_info())
        return None

cached_dictionary: Optional[BoqwizDictionary] = None

def load_dictionary() -> BoqwizDictionary:
    """
    Loads the currently installed version of the boQwI' dictionary.
    """

    global cached_dictionary
    if cached_dictionary:
        return cached_dictionary

    data = _try_load()
    if data:
        cached_dictionary = BoqwizDictionary.from_json(data)
        return cached_dictionary
    
    else:
        update_dictionary()
        return load_dictionary()

def update_dictionary():
    """
    Checks if there are available updates to the boQwI' dictionary and installs them.
    """

    global cached_dictionary

    # download the manifest and the latest version
    logger.info("Updating boQwI'...")
    try:
        manifest = requests.get(KAWHAQ_URL + "manifest.json").json()
        latest = manifest[FORMAT]["latest"]
    except:
        logger.error("Error while fetching the qawHaq manifest!", exc_info=sys.exc_info())
        return

    # check for existing versions of the dictionary

    if cached_dictionary and cached_dictionary.version == latest:
        logger.info(f"No update required.")
        return

    data = _try_load()
    if data and data["version"] == latest:
        logger.info(f"No update required.")
        return
    
    # install the update
    try:
        logger.info(f"Downloading boQwI' version {latest}...")
        url = KAWHAQ_URL + manifest[FORMAT][latest]["path"]
        archive = requests.get(url).content
        data = bz2.decompress(archive).decode("utf-8")

        with open(DICTIONARY_PATH, "w") as f:
            f.write(data)

        logger.info(f"Updated boQwI' to version {latest}.")
    
    except:
        logger.error("Error while updating the dictionary!", exc_info=sys.exc_info())
        return


from functools import reduce
import re
from typing import Callable, Dict, List, NamedTuple, Optional, Set, Tuple, Union
from yajwiz.types import Analysis, ProofreaderError, TokenType, Token

def _bitset_to_bitstring(bitset: Dict[str, str], included_bits: List[str]) -> str:
    return "".join(bitset.get(b, ".") for b in included_bits)

PUNCTUATION_NAMES = {
    ".": {"Period"},
    ",": {"Comma"},
    ":": {"Colon"},
    ";": {"Semicolon"},
    "!": {"ExclamationMark"},
    "?": {"QuestionMark"},
    "\"": {"QuotationMark"},
    "“": {"QuotationMark", "Left"},
    "”": {"QuotationMark", "Right"},
    "‘": {"QuotationMark", "Left"},
    "’": {"QuotationMark", "Right"},
    "«": {"QuotationMark", "Left"},
    "»": {"QuotationMark", "Right"},
    "‹": {"QuotationMark", "Left"},
    "›": {"QuotationMark", "Right"},
    "<": {"LessThanSign", "AngleBracket", "Left"},
    ">": {"GreaterThanSign", "AngleBracket", "Right"},
    "(": {"Parenthesis", "Left"},
    ")": {"Parenthesis", "Right"},
    "{": {"Brace", "Left"},
    "}": {"Brace", "Right"},
    "[": {"Bracket", "Left"},
    "]": {"Bracket", "Right"},
    "*": {"Asterisk"},
    "/": {"Slash"},
    "\\": {"Backslash"},
    "`": {"Backtick"},
    "~": {"Tilde"},
    "^": {"CircumflexMark"},
    "=": {"EqualsSign"},
    "+": {"PlusSign"},
    "-": {"MinusSign"},
    "–": {"Dash"},
    "_": {"Underscore"},
    "%": {"PercentSign"},
    "$": {"DollarSign", "CurrencySign"},
    "€": {"EuroSign", "CurrencySign"},
}

def _tokens_to_bitstring(tokens: List[Token], included_bits: List[str]) -> str:
    token_bitsets = []
    for token in tokens:
        if token.token_type == "PUNCT":
            bitset = {"Punct"}
            if token.text in PUNCTUATION_NAMES:
                bitset |= PUNCTUATION_NAMES[token.text]
            
            if len(token.text) == 1:
                bitset |= {str(ord(token.text))}
            
            token_bitsets.append([{bit: "1" for bit in bitset}])
        
        else:
            token_bitsets.append([a["SYNTAX_INFO"]["BITS"] for a in token.analyses])

    bitstring = ""
    for i, token_bitset in enumerate(token_bitsets):
        bitset = {}
        for bit in included_bits:
            if token_bitset and all(bit in s for s in token_bitset):
                bitset[bit] = "1"
            
            elif all(bit not in s for s in token_bitset):
                bitset[bit] = "0"
        
        bitstring += "," + str(i) + ":" + _bitset_to_bitstring(bitset, included_bits)
    
    return bitstring

def _pattern_to_regex(pattern: str) -> Tuple[re.Pattern, List[str]]:
    regex_parts = []
    while pattern:
        group_name = None
        if m := re.match(r"\w+", pattern):
            group_name = m.group(0)
            pattern = pattern[len(group_name):]
        
        if pattern[0] == "{":
            pattern = pattern[1:]
            end_index = pattern.index("}")
            bitcond = {}
            for bit in pattern[:end_index].split(","):
                val = "1"
                if bit.startswith("!"):
                    val = "0"
                    bit = bit[1:]
                
                bitcond[bit] = val
            
            if group_name:
                regex_parts += [f"(?P<{group_name}>"]
                regex_parts += [bitcond]
                regex_parts += [")"]
            
            else:
                regex_parts += [bitcond]
            
            pattern = pattern[end_index+1:]
        
        elif pattern[0] == "(":
            if group_name:
                regex_parts += [f"(?P<{group_name}>"]

            else:
                regex_parts += ["(?:"]
            
            pattern = pattern[1:]
        
        elif not group_name and pattern[0] in "*+?|)":
            regex_parts += [pattern[0]]
            pattern = pattern[1:]
        
        elif pattern[0] == "." or group_name and pattern[0].isspace():
            if group_name:
                regex_parts += [f"(?P<{group_name}>"]

            regex_parts += [{}]
            if group_name:
                regex_parts += [")"]
            
            pattern = pattern[1:]
        
        elif pattern[0].isspace():
            pattern = pattern[1:]
        
        else:
            raise Exception("Invalid pattern")
    
    bits = set()
    for part in regex_parts:
        if isinstance(part, dict):
            bits |= set(part.keys())
    
    included_bits = list(bits)

    regex = ""
    for part in regex_parts:
        if isinstance(part, str):
            regex += part
        
        elif isinstance(part, dict):
            regex += r",\d+:" + _bitset_to_bitstring(part, included_bits)
    
    return re.compile(regex), included_bits

class GrammarRule(NamedTuple):
    name: str
    pattern: str
    re_pattern: re.Pattern
    included_bits: List[str]
    message: str
    replacement: Optional[str]
    positive_examples: List[str]
    negative_examples: List[str]

    @staticmethod
    def from_dict(d={}, **d2) -> "GrammarRule":
        d = {**d, **d2}
        pattern, included_bits = _pattern_to_regex(d["pattern"])
        return GrammarRule(
            name=d["name"],
            pattern=d["pattern"],
            re_pattern=pattern,
            included_bits=included_bits,
            message=d["message"],
            replacement=d.get("replacement"),
            positive_examples=d.get("positive_examples", []),
            negative_examples=d.get("negative_examples", []),
        )

GRAMMAR_RULES: List[GrammarRule] = [
    GrammarRule.from_dict(
        name="aspect in complex sentence 1",
        pattern=r"({«'e'»}|{«net»}) ({«neH»}|{«neHHa'»})? {V7}",
        message="ASPECT SUFFIX IN COMPLEX SENTENCE",
        positive_examples=["DaH 'e' vIleghpu'", "DaH net leghtaH", "DaH 'e' neH vIleghpu'", "DaH net neHHa' leghtaH"],
        negative_examples=["DaH 'e' vIlegh", "DaH net legh", "DaH 'e' neH vIlegh", "DaH net neHHa' legh"],
    ),
    GrammarRule.from_dict(
        name="net with illegal subject",
        pattern=r"{«net»} ({Subj1}|{Subj2}|{-lu'})",
        message="net WITH ILLEGAL SUBJECT",
        positive_examples=["jISuv net vISov", "jISuv net boSov", "jISuv net Sovlu'"],
        negative_examples=["jISuv net Sov", "gheDlIj DaHoHHa'pu' moratlh\n'ej gheDDaj charghHa' molor\nQob qo' qeylIs yIqIm"],
    ),
    GrammarRule.from_dict(
        name="illegal 'e' or net",
        pattern=r"error({«'e'»}|{«net»}) verb({reH:v}|{Quj}|{jatlh:v}|{ja':v})",
        message="$error WITH $verb\\lemma",
        replacement="",
        positive_examples=["tugh jISuv 'e' qaja'."],
        negative_examples=["tugh jISuv. qaja'."]
    ),
    GrammarRule.from_dict(
        name="'e' with neH 1",
        pattern=r"error{«'e'»} {neH:v}",
        message="'e' WITH neH",
        replacement="",
        positive_examples=["tugh jISuv 'e' vIneH."],
        negative_examples=["tugh jISuv vIneH."]
    ),
    GrammarRule.from_dict(
        name="'e' with neH 2",
        pattern=r"error{«'e'»} {«neH»} {!Obj3}",
        message="'e' WITH neH",
        replacement="",
        positive_examples=["tugh jISuv 'e' neH qorDu'wIj."],
        negative_examples=["tugh jISuv neH qorDu'wIj.", "tugh jISuv 'e' neH DaSov."]
    ),
]

def _replace_groups(tokens: List[Token], groups: Dict[str, List[int]], text: str) -> str:
    for g in groups:
        if f"${g}" in text:
            text = text.replace(f"${g}\\lemma", " ".join(tokens[i].analyses[0]["LEMMA"] if tokens[i].analyses else tokens[i].text for i in groups[g]))
            text = text.replace(f"${g}", " ".join(tokens[i].text for i in groups[g]))
    
    return text

def proofread_tokens(tokens: List[Token], rules=GRAMMAR_RULES) -> List[ProofreaderError]:
    errors = []
    for token in tokens:
        if token.token_type == "WORD":
            if len(token.analyses) == 0:
                errors.append(ProofreaderError("unknown word", f"UNKNOWN WORD {token.text}", token.location, token.end_location()))
            
            elif all(a.get("UNGRAMMATICAL", None) for a in token.analyses):
                errors.append(ProofreaderError("ungrammatical", token.analyses[0]["UNGRAMMATICAL"], token.location, token.end_location()))
    
    for rule in rules:
        bitstring = _tokens_to_bitstring(tokens, rule.included_bits)
        #print(rule.included_bits)
        #print(rule.re_pattern)
        #print(bitstring)
        if m := rule.re_pattern.search(bitstring):
            groups = {key: [int(t[1:-1]) for t in re.findall(r",\d+:", value)] for key, value in m.groupdict().items()}
            #print(groups)

            replacement = rule.replacement
            if replacement is not None:
                replacement = _replace_groups(tokens, groups, replacement)
            
            message = _replace_groups(tokens, groups, rule.message)
            if "error" in groups:
                span = groups["error"]
            
            else:
                span = [int(t[1:-1]) for t in re.findall(r",\d+:", m.group(0))]
            
            start, end = tokens[span[0]].location, tokens[span[-1]].end_location()
            errors.append(ProofreaderError(rule.name, message, start, end, replacement=replacement))

    return errors

def test_grammar_rules():
    from . import analyzer
    succ = 0
    fail = 0
    for rule in GRAMMAR_RULES:
        examples: List[Tuple[str, str]] = [("Positive", example) for example in rule.positive_examples]
        examples += [("Negative", example) for example in rule.negative_examples]
        for example_type, example in examples:
            tokens = analyzer._tokenize_for_proofreader(example)
            errors = proofread_tokens(tokens, [rule])
            for error in errors:
                if example_type == "Positive" and rule.name == error.rule_name:
                    succ += 1
                
                elif rule.name == error.rule_name:
                    print(rule.name)
                    print(example_type + " example:", example)
                    print("->", error)
                    print()
                    fail += 1
            
            if not errors:
                if example_type == "Positive":
                    print(rule.name)
                    print("Positive example:", example)
                    print("NO ERRORS")
                    print()
                    fail += 1
                
                else:
                    succ += 1
    
    print(f"Result: {succ} ok, {fail} failed")

if __name__ == "__main__":
    test_grammar_rules()


from .analyzer import tokenize, split_to_morphemes, analyze, split_to_letters, split_to_syllables, get_errors
from .boqwiz import load_dictionary, update_dictionary, BoqwizDictionary, BoqwizEntry
from .pos_tagger import conllu_to_tagged_list, Tagger, text_to_conllu



from collections import defaultdict, Counter
import math
from re import T

from .analyzer import XPOS_INDEX, text_to_conllu_without_tagger, tokenize, analyze, _word_to_conllu

from typing import List, Tuple, Optional

def conllu_to_tagged_list(conllu_text: str) -> List[List[Tuple[str, Optional[str]]]]:
    sentences = []
    for line in conllu_text.split("\n"):
        line = line.strip()
        if not line:
            continue

        fields = line.split("\t")
        num = int(fields[0])
        form = fields[1]
        lemma = fields[2]
        _upos = fields[3]
        xpos = fields[4]

        if num == 1:
            sentences.append([])
        
        sentences[-1].append((lemma if lemma != "_" else form, xpos if xpos != "_" else None))
    
    return sentences

END_TOKEN: Tuple[str, Optional[str]] = ("$", "$")

class Tagger:
    """
    POS Tagger implementation that can fill missing POS information to sentences that have been mostly tagged
    """

    def __init__(self):
        self.word_dist = defaultdict(Counter)
        self.tag_dist1 = defaultdict(Counter)
        self.tag_dist2 = defaultdict(Counter)
        self.tags = set()
    
    def train(self, sentences: List[List[Tuple[str, Optional[str]]]]):
        for s in sentences:
            for (l1, p1), (_l2, p2), (_l3, p3) in zip(s, s[1:]+[END_TOKEN], s[2:]+[END_TOKEN, END_TOKEN]):
                if p1:
                    self.word_dist[p1][l1] += 1
                    if p2:
                        self.tag_dist1[p2][p1] += 1
                        self.tag_dist2[(p2, p3)][p1] += 1

                    self.tags.add(p1)
    
    def _get_word_prob(self, word, tag):
        p = self.word_dist[tag][word]
        t = sum(self.word_dist[tag].values())
        if p == 0 and word in XPOS_INDEX[tag] and tag in {"ADV", "CONJ"}:
            p = t / len(self.word_dist[tag])

        return -1000 if p == 0 else math.log(p / t)

    def _get_tag_prob(self, tag1, tag2, tag3):
        p = self.tag_dist2[(tag2, tag3)][tag1]
        t = sum(self.tag_dist2[(tag2, tag3)].values())
        if t == 0:
            p = self.tag_dist1[tag2][tag1]
            t = sum(self.tag_dist1[tag2].values())

        return -1000 if p == 0 else math.log(p / t)
    
    def tag(self, sent: List[Tuple[str, Optional[str]]]) -> List[Tuple[str, Optional[str]]]:
        sent = sent.copy()
        for i, ((l1, p1), (_l2, p2), (_l3, p3)) in reversed(list(enumerate(zip(sent, sent[1:] + [END_TOKEN], sent[2:] + [END_TOKEN, END_TOKEN])))):
            if not p1:
                max_score = -1e9
                max_tag: Optional[str] = None
                for tag in self.tags:
                    if "." in tag:
                        continue

                    score = self._get_word_prob(l1, tag) + self._get_tag_prob(tag, p2, p3)
                    if score > max_score:
                        max_score = score
                        max_tag = tag
                
                if max_tag:
                    sent[i] = (l1, max_tag)
        
        return sent

def text_to_conllu(text: str, tagger: Optional[Tagger] = None) -> str:
    """
    Converts a given text to the CONLL-U format with morphological information (dependencies are not parsed).
    If a word has multiple analyses, its POS and other info is not included (as they are not exactly known).

    If a tagger is provided, uses it to take the "best guess" when selecting from multiple analyses.
    """

    if not tagger:
        return text_to_conllu_without_tagger(text)
    
    ans = ""

    conllu = []
    tagged_sent = []
    tokens = tokenize(text)

    def tag_and_append():
        nonlocal ans, conllu, tagged_sent
        guessed_tags = tagger.tag(tagged_sent)
        for i, ((l1, p1), (_l2, p2)) in enumerate(zip(tagged_sent, guessed_tags)):
            if not p1 and p2:
                analyses = analyze(l1)
                for analysis in analyses:
                    if analysis["XPOS_GSUFF"] == p2:
                        conllu[i] = "\t".join(_word_to_conllu(i+1, l1, [analysis]))
                        break
        
        ans += "\n\n" + "\n".join(conllu)
        conllu = []
        tagged_sent = []

    i = 1
    for token_type, token in tokens:
        if token_type == "SPACE":
            continue

        elif token_type == "PUNCT":
            conllu.append("{}\t{}\t{}\tPUNCT\tPUNCT\t_\t_\t_\t_\t_".format(i, token, token))
            tagged_sent.append((token, "PUNCT"))
            if token in ".!?":
                i = 1
                tag_and_append()
            
            else:
                i += 1
        
        else:
            analyses = analyze(token)
            fields = _word_to_conllu(i, token, analyses)
            conllu.append("\t".join(fields))
            if fields[2] == "_":
                tagged_sent.append((token, None))
            
            else:
                tagged_sent.append((fields[2], fields[4]))
            i += 1
    
    if conllu:
        tag_and_append()
    
    return ans


from typing import Dict, List, Literal, NamedTuple, Optional, Set, Tuple, TypedDict
from .tables import Person, Number

# Morphological analysis

Xpos = Literal["VS", "VT", "VI", "VA", "V?", "NL", "NB", "PRON", "NUM", "N", "ADV", "EXCL", "CONJ", "QUES", "UNK"]

class SyntaxInfo(TypedDict, total=False):
    ROLE: Literal["NP", "VP", "OTHER"]
    OBJECT_PERSON: Set[Person]
    OBJECT_NUMBER: Optional[Number]
    SUBJECT_PERSON: Set[Person]
    SUBJECT_NUMBER: Optional[Number]
    PLURAL: bool
    BITS: Set[str]

class _AnalysisRequired(TypedDict):
    WORD: str
    LEMMA: str
    POS: str
    XPOS: Xpos
    BOQWIZ_POS: str
    BOQWIZ_ID: str
    PARTS: List[str]

class Analysis(_AnalysisRequired, total=False):
    PREFIX: str
    SUFFIX: Dict[str, str]
    XPOS_GSUFF: str
    UNGRAMMATICAL: str
    SYNTAX_INFO: SyntaxInfo

TokenType = Literal["WORD", "SPACE", "PUNCT"]

class Token(NamedTuple):
    location: int
    token_type: TokenType
    text: str
    analyses: List[Analysis]

    def end_location(self):
        return self.location + len(self.text)

# Grammar checking

class ProofreaderError(NamedTuple):
    rule_name: str
    message: str
    location: int
    end_location: int
    replacement: Optional[str] = None


import re
from collections import defaultdict
import copy

from typing import DefaultDict, Dict, List, NamedTuple, Set, Tuple, Optional, Literal, TypedDict
from yajwiz.grammar_rules import proofread_tokens
from yajwiz.boqwiz import BoqwizEntry, load_dictionary

from .tables import SUFFIX_TYPES, UNIVERSAL_FEATURES, XPOS_TO_UPOS, PREFIX_TABLE, Person, Number
from .types import ProofreaderError, Token, TokenType, Xpos, Analysis, SyntaxInfo

def _get_xpos(entry: BoqwizEntry) -> Xpos:
    pos2 = entry.tags
    if "v" in pos2 and "is" in pos2:
        return "VS"
    
    elif "v" in pos2 and pos2 & {"t", "t_c"}:
        return "VT"
    
    elif "v" in pos2 and pos2 & {"i", "i_c"}:
        return "VI"
    
    elif pos2 & {"ambi"}:
        return "VA"
    
    elif pos2 & {"v"}:
        return "V?"
    
    elif pos2 & {"n"} and pos2 & {"being"}:
        return "NL"
    
    elif pos2 & {"n"} and pos2 & {"body"}:
        return "NB"
    
    elif pos2 & {"n"} and ("pro" in pos2 or entry.name in {"'Iv", "nuq", "jIH"}):
        return "PRON"
    
    elif pos2 & {"n"} and entry.name in {"wa'", "cha'", "wej", "loS", "vagh", "jav", "Soch", "chorgh", "Hut"}:
        return "NUM"
    
    elif pos2 & {"n"}:
        return "N"
    
    elif pos2 & {"adv"}:
        return "ADV"
    
    elif pos2 & {"excl"}:
        return "EXCL"
    
    elif pos2 & {"conj"}:
        return "CONJ"
    
    elif pos2 & {"ques"}:
        return "QUES"
    
    else:
        return "UNK"

dictionary = load_dictionary()

VERBS: List[str] = []
STATIVE_VERBS: List[str] = ["lo'laH", "lo'laHbe'"]
NOUNS: List[str] = []

DERIV_VERBS: List[str] = ["lo'laH", "lo'laHbe'"]
DERIV_STATIVE_VERBS: List[str] = []
DERIV_NOUNS: List[str] = []

ALL_WORDS: Set[str] = set()
WORD_INDEX: DefaultDict[str, List[BoqwizEntry]] = defaultdict(lambda: [])
XPOS_INDEX: DefaultDict[str, Set[str]] = defaultdict(set)

for boqwiz_id in dictionary.entries:
    entry = dictionary.entries[boqwiz_id]
    word = entry.name
    pos = entry.tags
    if "hyp" in pos:
        continue

    very_bad = "pref" in pos or "suff" in pos

    if "v" in pos:
        good = not very_bad and word not in DERIV_VERBS and word not in DERIV_STATIVE_VERBS
        if good:
            VERBS.append(word)
        
        elif not very_bad:
            DERIV_VERBS.append(word)

        WORD_INDEX[word + ":v"].append(entry)
    
        if good and "is" in pos:
            STATIVE_VERBS.append(word)
        
        elif not very_bad:
            DERIV_STATIVE_VERBS.append(word)
    
    elif "n" in pos:
        if not very_bad and word not in DERIV_NOUNS:
            NOUNS.append(word)
        
        elif not very_bad:
            DERIV_NOUNS.append(word)

        WORD_INDEX[word + ":n"].append(entry)
    
    else:
        WORD_INDEX[word + ":other"].append(entry)
    
    ALL_WORDS.add(word)
    XPOS_INDEX[_get_xpos(entry)].add(word)

# Match longer first
VERBS.sort(key=lambda i: -len(i))
STATIVE_VERBS.sort(key=lambda i: -len(i))
NOUNS.sort(key=lambda i: -len(i))

NOUN_SUFFIX_REGEX = r"('a'|Hom|(?<=[bDHjqlmnpQrStvwy'hg])oy|(?<![bDHjqlmnpQrStvwy'hg])'oy)?(pu'|Du'|mey)?(qoq|Hey|na')?(wI'|ma'|lI'|ra'|wIj|maj|lIj|raj|Daj|chaj|vam|vetlh)?(Daq|vo'|mo'|vaD|'e')?"

NOUN_REGEX = re.compile(r"")

VERB_SUFFIX_REGEX = r"(Ha')?(be'|qu')?('egh|chuq)?(be'|qu')?(nIS|qang|rup|beH|vIp)?(be'|qu')?(choH|qa')?(be'|qu')?(moH)?(be'|qu')?(lu'|laH)?(be'|qu')?(chu'|bej|ba'|law')?(be'|qu')?(pu'|ta'|taH|lI')?(be'|qu')?(neS)?(be'|qu')?(Qo')?(?:(DI'|chugh|pa'|vIS|mo'|bogh|meH|'a'|jaj)|(?:(wI'|ghach)" + NOUN_SUFFIX_REGEX + r"))?"

VERB_REGEX = re.compile(r"")

PRONOUN_VERB_REGEX = re.compile(r"(jIH|maH|SoH|tlhIH|ghaH|chaH|'oH|bIH)" + VERB_SUFFIX_REGEX)

STATIVE_VERB_REGEX = re.compile(r"")

NUMBER_SUFFIX_REGEX = r"(maH|vatlh|SaD|SanID|netlh|bIp|'uy'|Saghan|maH'uy'|vatlhbIp|vatlh'uy'|SaDbIp|SanIDbIp)"
NUMBER_REGEX = re.compile(r"(wa'|cha'|wej|loS|vagh|jav|Soch|chorgh|Hut)(?:" + NUMBER_SUFFIX_REGEX +  r"(DIch|logh|leS|Hu')?|" + NUMBER_SUFFIX_REGEX + r"?(DIch|logh|leS|Hu'))")

def _create_regexes():
    global NOUN_REGEX, VERB_REGEX, STATIVE_VERB_REGEX

    NOUN_REGEX = re.compile(r"(" + r"|".join(NOUNS) + r")" + NOUN_SUFFIX_REGEX)

    VERB_REGEX = re.compile(r"(|HI|gho|yI|tI|pe|qa|Sa|vI|jI|pI|re|DI|wI|ma|cho|ju|Da|bI|tu|che|bo|Su|mu|nu|Du|lI|nI|lu)("
        + r"|".join(VERBS)
        + r")" + VERB_SUFFIX_REGEX)

    STATIVE_VERB_REGEX = re.compile(r"(" + r"|".join(STATIVE_VERBS) + r")(Ha')?(be')?(qu')?(be')?(Daq|vo'|mo'|vaD|'e')")

# Find derived words that don't mess with parsing and add them to the regexes

def _add_if_does_not_match(derived: List[str], all: List[str]):
    for word in derived:
        if not any(regex.fullmatch(word) for regex in [NOUN_REGEX, STATIVE_VERB_REGEX, NUMBER_REGEX, PRONOUN_VERB_REGEX, VERB_REGEX]):
            all.append(word)

_create_regexes() # Create regexes for the first time
_add_if_does_not_match(DERIV_VERBS, VERBS)
_add_if_does_not_match(DERIV_STATIVE_VERBS, STATIVE_VERBS)
_add_if_does_not_match(DERIV_NOUNS, NOUNS)
_create_regexes() # Create regexes for the second time with the additional words

def split_to_morphemes(word: str) -> Set[tuple]:
    """
    Given a word, splits it to morphemes. Prefixes and suffixes are marked with dashes.
    """
    ans = set()
    for regex in [NOUN_REGEX, STATIVE_VERB_REGEX, NUMBER_REGEX, PRONOUN_VERB_REGEX]:
        if m := regex.fullmatch(word):
            parts = []
            for i, part in enumerate(m.groups()):
                if not part:
                    continue

                if i > 0:
                    part = "-" + part

                parts.append(part)
            
            ans.add(tuple(parts))
    
    if m := VERB_REGEX.fullmatch(word):
        parts = []
        for i, part in enumerate(m.groups()):
            if not part:
                continue

            if i < 1:
                part = part + "-"

            elif i > 1:
                part = "-" + part

            parts.append(part)
        
        ans.add(tuple(parts))
    
    return ans

def _next_morphem(parsed: List[str], i: int) -> Tuple[int, Optional[str]]:
    oi = i
    i += 1
    while i < len(parsed):
        if parsed[i]:
            return i, parsed[i]

        i += 1
    
    return oi, None

LETTER = re.compile(r"(ch|gh|ng|tlh|[abDeHIjlmnopqQrStuvwy'])")

def split_to_letters(word: str) -> List[str]:
    """
    Splits the given word to Klingon letters. The word must not contain spaces or foreign letters.
    """
    letters = []
    while word:
        if m := re.match(LETTER, word):
            letter = m.group(1)
            word = word[len(letter):]
            letters += [letter]

        else:
            word = word[1:]

    return letters

def split_to_syllables(word: str) -> List[str]:
    """
    Splits the given word to syllables. The word msut not contain spaces or foreign words and it should follow the Klingon phonotactics.
    """
    letters = split_to_letters(word)
    syllables = [""]

    for prev, cur, succ in zip([""]+letters, letters, letters[1:]+[""]):
        if syllables[-1] and \
            cur not in "aeIou" and \
            (succ and succ in "aeIou" or \
                prev and prev not in "aeIou" and \
                not (prev+cur) in {"w'", "y'", "rgh"}):
            syllables.append(cur)

        else:
            syllables[-1] += cur
    
    return syllables

def _analyze_word_with_pos(ans: List[Analysis], start_pos: str, regex: re.Pattern, lemma_idx: int, word: str, infl_pos:str=None, lemma_pred=lambda l: True):
    if m := regex.fullmatch(word):
        parsed = list(m.groups())
        def rec(i: int, pos: str, obj: Analysis):
            if i >= len(parsed):
                return [obj]

            part = parsed[i]
            if not part:
                return rec(i + 1, pos, obj)

            if i < lemma_idx:
                part = part + "-"
                obj["PREFIX"] = part

            elif i > lemma_idx:
                part = "-" + part
            
                if pos == "v" and part in {"-Daq", "-vo'", "-mo'", "-vaD", "-'e'"}:
                    pos = "n"

                if "SUFFIX" not in obj:
                    obj["SUFFIX"] = {}
                
                obj["SUFFIX"][SUFFIX_TYPES[(part, pos)]] = part
                j, m = _next_morphem(parsed, i)
                if m in {"be'", "qu'"}:
                    obj["SUFFIX"][SUFFIX_TYPES[(part, pos)]] += m
                    i = j
            
            else:
                obj["LEMMA"] = part

            new_pos = pos

            if pos == "v" and part in ["-ghach", "-wI'"]:
                new_pos = "n"
            
            if i == lemma_idx and infl_pos:
                new_pos = infl_pos
            
            if (part + ":" + pos) in WORD_INDEX:
                objs = []
                for entry in WORD_INDEX[part + ":" + pos]:
                    new_obj = copy.deepcopy(obj)
                    new_obj["PARTS"].append(entry.id)
                    if i == lemma_idx:
                        #new_obj["BOQWIZ"] = entry

                        new_obj["XPOS"] = _get_xpos(entry)
                        new_obj["BOQWIZ_POS"] = entry.part_of_speech
                        new_obj["BOQWIZ_ID"] = entry.id

                        if not lemma_pred(entry):
                            continue
                    
                    objs += rec(i + 1, new_pos, new_obj)
                
                return objs
            
            else:
                obj["PARTS"].append(part)
            
                return rec(i + 1, new_pos, obj)
        
        start_obj: Analysis = {
            "WORD": word,
            "POS": start_pos.upper(),
            "XPOS": "UNK",
            "BOQWIZ_POS": "?",
            "BOQWIZ_ID": "?",
            "PARTS": [],
            "LEMMA": "",
        }

        objs = rec(0, start_pos, start_obj)
        
        ans += objs

GENDERED_SUFFIXES = {
    "being": {
        "-wI':n",
        "-lI':n",
        "-ma':n",
        "-ra':n",
        "-pu':n",
    },
    "body": {"-Du':n"},
    "other": set(),
}

def analyze(word: str, include_syntactical_info=False) -> List[Analysis]:
    """
    Given a word, returns a list of possible analyses.

    Each analysis has:
    - WORD: the analysed word itself
    - LEMMA: the base form of the word, without any affixes
    - POS: either `"N"`, `"V"` or `"OTHER"`
    - XPOS: a more detailed part of speech tag
    - XPOS_GSUFF: XPOS + grammatical suffix (N5 or V9 type suffix or grammatical number suffix)
    - PARTS: a list of boQwI' identifiers for morphemes
    - PREFIX: (optional) the prefix of the word
    - SUFFIX: (optional) a dict where key is the suffix type (like V7) and values is the suffix (like `-ta'be'`). Rovers are included in their preceding suffixes
    """
    ans: List[Analysis] = []
    
    _analyze_word_with_pos(ans, "n", NOUN_REGEX, 0, word)
    _analyze_word_with_pos(ans, "n", NUMBER_REGEX, 0, word)
    if not ans or len(ans[0]["PARTS"]) > 1:
        _analyze_word_with_pos(ans, "n", PRONOUN_VERB_REGEX, 0, word, infl_pos="v", lemma_pred=lambda e: "pro" in e.tags)

    _analyze_word_with_pos(ans, "v", VERB_REGEX, 1, word)
    _analyze_word_with_pos(ans, "v", STATIVE_VERB_REGEX, 0, word, lemma_pred=lambda e: "is" in e.tags)

    if word + ":other" in WORD_INDEX:
        for entry in WORD_INDEX[word + ":other"]:
            ans.append({
                "WORD": word,
                "LEMMA": entry.name,
                "POS": "OTHER",
                "XPOS": _get_xpos(entry),
                "BOQWIZ_POS": entry.part_of_speech,
                "BOQWIZ_ID": entry.id,
                "PARTS": [ entry.id ],
            })
    
    for analysis in ans:
        analysis["XPOS_GSUFF"] = analysis["XPOS"]
        if "SUFFIX" in analysis:
            if "V9" in analysis["SUFFIX"]:
                analysis["XPOS_GSUFF"] += "." + analysis["SUFFIX"]["V9"][1:]
            
            if "N5" in analysis["SUFFIX"]:
                analysis["XPOS_GSUFF"] += "." + analysis["SUFFIX"]["N5"][1:]
            
            if "L2" in analysis["SUFFIX"]:
                analysis["XPOS_GSUFF"] += "." + analysis["SUFFIX"]["L2"][1:]
        
        # Check for easy morphological errors:
        
        if analysis["POS"] == "N" and analysis["LEMMA"] not in {"qor", "chuD"}:
            gender = "body" if "body" in analysis["BOQWIZ_POS"] else \
                "being" if "being" in analysis["BOQWIZ_POS"] or "name" in analysis["BOQWIZ_POS"] else \
                "other"

            if analysis["LEMMA"] in {"qorDu'", "latlh"}: # these can be either language users or others
                gender = "being"
            
            for part in analysis["PARTS"]:
                for other_gender in {"being", "body", "other"} - {gender}:
                    if part in GENDERED_SUFFIXES[other_gender]:
                        analysis["UNGRAMMATICAL"] = "ILLEGAL PLURAL OR POSSESSIVE SUFFIX"
        
        if "-vIS:v" in analysis["PARTS"] and \
            "-taH:v" not in analysis["PARTS"]:
            analysis["UNGRAMMATICAL"] = "-vIS WITHOUT -taH"
        
        if "-lu':v" in analysis["PARTS"] and analysis.get("PREFIX", "") not in {"vI-", "Da-", "wI-", "bo-", "", "lu-"}:
            analysis["UNGRAMMATICAL"] = "ILLEGAL PREFIX WITH -lu'"
        
        # Do not include XPOS=VI here because data is unreliable
        if analysis["XPOS"] in {"VS"} and analysis.get("PREFIX", "") not in {"yI-", "pe-", "jI-", "bI-", "ma-", "Su-", ""} and "-moH:v" not in analysis["PARTS"]:
            analysis["UNGRAMMATICAL"] = "ILLEGAL SUFFIX WITH INTRANSITIVE VERB"
        
        if "-ghach:v" in analysis["PARTS"] and analysis["PARTS"].index("-ghach:v") - analysis["PARTS"].index(analysis["BOQWIZ_ID"]) < 2 and analysis["LEMMA"] not in {"lo'laH", "lo'laHbe'"}:
            analysis["UNGRAMMATICAL"] = "-ghach WITHOUT OTHER SUFFIX"

        if "-jaj:v" in analysis["PARTS"] and analysis.get("SUFFIX", {}).get("V7", "") != "":
            analysis["UNGRAMMATICAL"] = "-jaj WITH ASPECT"
        
        # Add extra information regarding the words rule in the syntax
        if include_syntactical_info:
            bits = {analysis["XPOS"], analysis["POS"], f"«{analysis['WORD']}»"}
            info: SyntaxInfo = {}
            analysis["SYNTAX_INFO"] = info
            if analysis["POS"] == "N":
                info["ROLE"] = "NP"
            
            elif analysis["POS"] == "V" and analysis.get("SUFFIX", {}).get("V9", None) in {"-wI'", "-ghach"}:
                info["ROLE"] = "NP"
            
            elif analysis["POS"] == "V":
                info["ROLE"] = "VP"
            
            else:
                info["ROLE"] = "OTHER"
            
            bits |= {info["ROLE"]}
            
            if info["ROLE"] == "VP":
                voice = "NP" if "-lu':v" in analysis["PARTS"] else "P"
                subj_person: Set[Person]
                subj_number: Optional[Number]
                obj_person: Set[Person]
                obj_number: Optional[Number]
                if "PREFIX" in analysis or voice == "NP":
                    if (analysis.get("PREFIX", "-"), voice) not in PREFIX_TABLE: # ungrammatical word
                        subj_person = set()
                        subj_number = None
                        obj_person = set()
                        obj_number = None
                    
                    else:
                        subj_person, subj_number, obj_person, obj_number = PREFIX_TABLE[(analysis.get("PREFIX", "-"), voice)]
                
                elif analysis["XPOS"] in {"VS", "VI"}:
                    subj_person = {3}
                    subj_number = None
                    obj_person = {0}
                    obj_number = None

                else:
                    subj_person = {3}
                    subj_number = None
                    obj_person = {0, 3}
                    obj_number = None
                
                info["SUBJECT_PERSON"] = subj_person
                info["SUBJECT_NUMBER"] = subj_number
                info["OBJECT_PERSON"] = obj_person
                info["OBJECT_NUMBER"] = obj_number

                for person in subj_person:
                    bits |= {f"Subj{person}{subj_number or ''}", f"Subj{person}"}
                    if subj_number:
                        bits |= {f"Subj{subj_number}"}

                for person in obj_person:
                    bits |= {f"Obj{person}{subj_number or ''}", f"Obj{person}"}
                    if obj_number:
                        bits |= {f"Obj{obj_number}"}
            
            elif info["ROLE"] == "NP":
                if "inhps" in analysis["BOQWIZ_POS"] or "inhpl" in analysis["BOQWIZ_POS"]:
                    info["PLURAL"] = False
                    bits |= {"Singular"}

                if analysis.get("SUFFIX", {}).get("N2", None) in {"-pu'", "-Du'", "-mey"}:
                    info["PLURAL"] = True
                    bits |= {"Plural"}
            
            
            for part in analysis["PARTS"]:
                if ":" not in part:
                    bits |= {part}
                
                else:
                    bits |= {part, part[:part.index(":")]}
            
            for key, val in analysis.get("SUFFIX", {}).items():
                bits |= {key, val}
            
            if prefix := analysis.get("PREFIX", None):
                bits |= {prefix}
            
            info["BITS"] = bits

    return ans

def _word_to_conllu(num: int, word: str, analyses: List[Analysis]) -> tuple:
    if len(analyses) == 0:
        return (str(num), word, "_", "_", "_", "_", "_", "_", "_", "_")
    
    if len(analyses) > 1:
        if (all(a["POS"] == analyses[0]["POS"] for a in analyses)
            and all(a["LEMMA"] == analyses[0]["LEMMA"] for a in analyses)):
            pass

        else:
            return (str(num), word, "_", "_", "_", "_", "_", "_", "_", "_")
    
    analysis = analyses[0]
    feats = []
    extra = []
    if "PREFIX" in analysis:
        extra.append("Prefix=" + analysis["PREFIX"])
        f = UNIVERSAL_FEATURES.get((analysis["PREFIX"], "P" if not analysis.get("SUFFIX", {}).get("V5", "") == "-lu'" else "NP"), None)
        if f:
            feats.append(f)
    
    elif XPOS_TO_UPOS[analysis["XPOS"]] in {"VERB", "ADJ"}:
        f = UNIVERSAL_FEATURES.get(("-", "P" if not analysis.get("SUFFIX", {}).get("V5", "") == "-lu'" else "NP"), None)
        if f:
            feats.append(f)
    
    xpos = analysis["XPOS_GSUFF"]
    upos = XPOS_TO_UPOS[analysis["XPOS"]]
    
    if "SUFFIX" in analysis:
        for key in analysis["SUFFIX"]:
            extra.append("Suffix" + key + "=" + analysis["SUFFIX"][key])
            if (analysis["SUFFIX"][key], key) in UNIVERSAL_FEATURES:
                feats.append(UNIVERSAL_FEATURES[(analysis["SUFFIX"][key], key)])

    return (str(num), word, analysis["LEMMA"], upos, xpos, "_" if not feats else "|".join(feats), "_", "_", "_", "_" if not extra else "|".join(extra))

def tokenize(sentence: str) -> List[Tuple[TokenType, str]]:
    """
    Tokenizes the given text and returns a list of tuples with form `(type, value)` where type is one of `"WORD"`, `"SPACE"`, `"PUNCT"`.
    """
    tokens: List[Tuple[TokenType, str]] = []
    for matches in re.findall(r"([a-zA-Z'0-9]+|\s+|(.)\2*)", sentence):
        token = matches[0]
        if re.fullmatch(r"[a-zA-Z'0-9]+", token):
            tokens.append(("WORD", token))
        
        elif re.fullmatch(r"\s+", token):
            tokens.append(("SPACE", token))
        
        elif token:
            tokens.append(("PUNCT", token))
    
    return tokens

def _tokenize_for_proofreader(sentence: str) -> List[Token]:
    tokens: List[Token] = []
    char = 0
    for token in tokenize(sentence):
        if token[0] != "SPACE":
            tokens.append(Token(char, token[0], token[1], analyze(token[1], include_syntactical_info=True)))
        
        char += len(token[1])
    
    return tokens

def get_errors(sentence: str) -> List[ProofreaderError]:
    tokens = _tokenize_for_proofreader(sentence)
    return proofread_tokens(tokens)

def get_errors_old(sentence: str) -> List[ProofreaderError]:
    errors: List[ProofreaderError] = []
    tokens: List[Tuple[int, TokenType, str]] = []
    char = 0
    for token in tokenize(sentence):
        if token[0] != "SPACE":
            tokens.append((char, token[0], token[1]))
        
        char += len(token[1])
    
    for i in range(len(tokens)):
        pos, token1_type, token1_text = tokens[i]
        end_pos = pos + len(token1_text)
        if token1_type == "WORD":
            token1 = analyze(token1_text)
            if not token1:
                errors.append(ProofreaderError("unknown word", f"UNKNOWN WORD {token1_text}", pos, end_pos))
                continue

            if all(token.get("UNGRAMMATICAL", None) for token in token1):
                errors.append(ProofreaderError("ungrammatical", token1[0]["UNGRAMMATICAL"], pos, end_pos))

            if i > len(tokens)-2:
                continue

            token2_pos, token2_type, token2_text = tokens[i+1]
            token2_end_pos = token2_pos + len(token2_text)

            if not token2_type or token2_type != "WORD":
                continue

            token2 = analyze(token2_text)
            if not token2:
                continue

            if token1_text == "net":
                if all("V7" in token.get("SUFFIX", {}) for token in token2):
                    errors.append(ProofreaderError("", "ASPECT SUFFIX IN COMPLEX SENTENCE", pos, token2_end_pos))

            elif token1_text == "'e'":
                if all("V7" in token.get("SUFFIX", {}) for token in token2):
                    errors.append(ProofreaderError("", "ASPECT SUFFIX IN COMPLEX SENTENCE", pos, token2_end_pos))

                token3_pos, token3_type, token3_text = tokens[i+2] if i <= len(tokens)-3 else (0, None, "")
                token3_end_pos = token3_pos + len(token3_text)

                if token3_type and token3_type == "WORD":
                    token3 = analyze(token3_text, True)

                else:
                    token3 = []

                if token2_text in {"neH", "neHHa'", "je"} and token3 and all("V7" in token.get("SUFFIX", {}) for token in token3):
                    errors.append(ProofreaderError("", "ASPECT SUFFIX IN COMPLEX SENTENCE", pos, token3_end_pos))

                if all(token["BOQWIZ_ID"] == "neH:v" for token in token2) or token2_text == "neH" and not any(3 in token.get("SYNTAX_INFO", {}).get("OBJECT_PERSON", set()) for token in token3):
                    errors.append(ProofreaderError("", "'e' WITH neH", pos, token2_end_pos))

    return errors

def text_to_conllu_without_tagger(text: str) -> str:
    """
    Converts a given text to the CONLL-U format with morphological information (dependencies are not parsed).
    If a word has multiple analyses, its POS and other info is not included (as they are not exactly known).
    """
    ans = []
    tokens = tokenize(text)
    i = 1
    for token_type, token in tokens:
        if token_type == "SPACE":
            continue

        elif token_type == "PUNCT":
            ans.append("{}\t{}\t{}\tPUNCT\tPUNCT\t_\t_\t_\t_\t_".format(i, token, token))
            if token in ".!?":
                i = 1
                ans.append("")
            
            else:
                i += 1
        
        else:
            analyses = analyze(token)
            ans.append("\t".join(_word_to_conllu(i, token, analyses)))
            i += 1
    
    return "\n".join(ans)



import argparse
import re
import sys

import yajwiz

def main():
    parser = argparse.ArgumentParser(description="Klingon grammar checker")
    parser.add_argument("input_file", default="-", help="The text file to be processed")
    parser.add_argument("-I", "--ignore_unknown", action="store_true", help="Ignore unknown word errors")
    parser.add_argument("-w", "--additional_words", help="A comma-separated list of additional words to be added to the dictionary")
    parser.add_argument("-W", "--additional_words_file", help="A file that contains one additional word per line to be added to the dictionary")
    args = parser.parse_args()

    words = set()
    if args.additional_words:
        words |= {"UNKNOWN WORD "+word.strip() for word in args.additional_words.split(",")}
    
    if args.additional_words_file:
        with open(args.additional_words_file, "r") as f:
            for line in f:
                line = line.strip()
                if line:
                    words.add("UNKNOWN WORD "+line)

    if args.input_file and args.input_file != "-":
        input_file = open(args.input_file, "r")
    
    else:
        input_file = sys.stdin
    
    count = 0
    for i, line in enumerate(input_file):
        line = line.strip()
        errors = yajwiz.get_errors(line)
        sentences = re.split(r"[.!?] ", line)
        j = 0
        for error in errors:
            if error.message in words or args.ignore_unknown and error.message.startswith("UNKNOWN WORD"):
                continue
            
            while j+len(sentences[0]) < error.location:
                j += len(sentences[0]) + 2
                sentences.pop(0)
            
            prefix = "Line " + str(i+1) + ": "
            print(prefix + sentences[0])
            print(prefix + " "*(error.location - j) + "^" + error.message)
            print()
            count += 1
    
    print(f"Found {count} errors")
    
    input_file.close()

if __name__ == "__main__":
    main()


from setuptools import setup

setup(
    name="yajwiz",
    version="0.9",
    author="Iikka Hauhio",
    author_email="fergusq@kaivos.org",
    packages=["yajwiz"],
    description="Klingon NLP toolkit",
    long_description=open("README.rst").read(),
    classifiers=[
        "Programming Language :: Python :: 3"
    ],
    python_requires='>=3.8',
    install_requires=["appdirs", "requests"]
)
